use rand::prelude::ThreadRng;
use rand::Rng;

use super::{StochasticBehavior, StopCondition};
use std::f64::consts::PI;

const DEFAULT_PERIOD: f64 = 1000.0;
const MIN_RAND_UPPER_BOUND: f64 = 1.25;

pub struct RandM {
    stochastic_behavior: StochasticBehavior,
    num_directions: usize,
    num_variables: usize,
    period: f64,
}

impl RandM {
    pub fn new(
        stochastic_behavior: StochasticBehavior,
        stop_condition: &StopCondition,
        num_variables: usize,
        num_directions: usize,
    ) -> Self {
        let period: f64 = match stop_condition {
            StopCondition::Iterations(iters) => *iters as f64,
            StopCondition::Cost(_) => DEFAULT_PERIOD,
            StopCondition::Both(iters, _) => *iters as f64,
        };

        Self {
            stochastic_behavior,
            num_directions,
            num_variables,
            period,
        }
    }

    pub fn rand_vecs(&self, iter: usize, rng: &mut ThreadRng) -> Vec<Vec<f64>> {
        let progress = (iter as f64) / self.period;

        let r_upper_bounds: Vec<f64> = (0..self.num_directions)
            .map(|d| {
                self.stochastic_behavior.r_max
                    * ((PI * progress * ((self.stochastic_behavior.r_freq + 2 * d) as f64)
                        + self.stochastic_behavior.r_phase)
                        .cos()
                        + MIN_RAND_UPPER_BOUND)
            })
            .collect();

        (0..self.num_directions)
            .map(|d| {
                (0..self.num_variables)
                    .map(|_| rng.gen_range(self.stochastic_behavior.r_min, r_upper_bounds[d]))
                    .collect()
            })
            .collect()
    }

    pub fn momentum(&self, iter: usize) -> f64 {
        let progress = (iter as f64) / self.period;

        self.stochastic_behavior.m_mag
            * (PI * progress * (self.stochastic_behavior.m_freq as f64)
                + self.stochastic_behavior.m_phase)
                .cos()
    }
}
