use std::fmt;

const DEFAULT_WALL_BOUNCE: f64 = 0.125;
const DEFAULT_LOCAL: f64 = 0.8;
const DEFAULT_TRIBAL: f64 = 1.2;
const DEFAULT_MOMENTUM: f64 = 0.5;
const DEFAULT_NUM_PARTICLES: usize = 128;

/// Configuration for Swarms of Particles
#[derive(Clone, Debug)]
pub struct SwarmConfig {
    pub global_behavior: GlobalBehavior,
    pub wall_bounce_factor: f64,
    pub local_motion: f64,
    pub tribal_motion: f64,
    pub momentum: f64,
    pub num_particles: usize,
    pub stochastic_behavior: StochasticBehavior,
    pub update_console: Option<usize>,
}

impl SwarmConfig {
    pub fn new() -> Self {
        Self {
            global_behavior: GlobalBehavior::default(),
            wall_bounce_factor: DEFAULT_WALL_BOUNCE,
            local_motion: DEFAULT_LOCAL,
            tribal_motion: DEFAULT_TRIBAL,
            momentum: DEFAULT_MOMENTUM,
            num_particles: DEFAULT_NUM_PARTICLES,
            stochastic_behavior: StochasticBehavior::new(),
            update_console: Some(100),
        }
    }

    pub fn synergic_behavior(
        &mut self,
        global_motion_coefficient: f64,
        collaboration_periodicity: usize,
    ) {
        self.global_behavior =
            GlobalBehavior::Synergic(global_motion_coefficient, collaboration_periodicity);
    }

    pub fn solitary_behavior(&mut self) {
        self.global_behavior = GlobalBehavior::Solitary;
    }

    pub fn wall_bounce(&mut self, wall_bounce_factor: f64) {
        assert!(
            wall_bounce_factor > 0.0,
            "Wall Bounce Factor must be greater than zero!"
        );
    }

    pub fn motion_coefficients(&mut self, local_motion: f64, tribal_motion: f64, momentum: f64) {
        self.local_motion = local_motion;
        self.tribal_motion = tribal_motion;
        self.momentum = momentum;
    }

    pub fn num_particles(&mut self, num_particles: usize) {
        self.num_particles = num_particles;
    }

    pub fn set_stochastic_behavior(&mut self, stochastic_behavior: StochasticBehavior) {
        assert!(
            self.momentum - (stochastic_behavior.m_mag / 2.0) > 0.0,
            "Momentum Magnitde must be greater than half of the Momentum!"
        );

        self.stochastic_behavior = stochastic_behavior;
    }
}

impl fmt::Display for SwarmConfig {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{} Particles: ", self.num_particles)?;
        write!(f, "Particle motion coefficients: \n")?;
        write!(
            f,
            "\t [Local: {}, tribal: {}] \t momentum: {} \n \t wall_bounce_factor: {} \n",
            self.local_motion, self.tribal_motion, self.momentum, self.wall_bounce_factor,
        )?;
        write!(f, "{} \n", self.global_behavior)?;
        write!(f, "{}", self.stochastic_behavior)
    }
}

#[derive(Clone, Debug)]
pub enum GlobalBehavior {
    /// collaboration with: (global motion coefficient, collaboration rate)
    Synergic(f64, usize),
    /// no inter-swarm collaboration
    Solitary,
}

impl GlobalBehavior {
    pub fn default() -> Self {
        Self::Solitary
    }
}

impl fmt::Display for GlobalBehavior {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "Swarm Behavior: ")?;
        match self {
            Self::Synergic(global_motion_coeff, collaboration) => {
                write!(
                    f,
                    "Swarm will collaborate with others every {} itterations \n and weight the global record with a coefficient of {}.",
                    collaboration, global_motion_coeff
                )?;
            }
            Self::Solitary => {
                write!(f, "Swarm will not collaborate with others. \n")?;
            }
        }

        write!(f, "")
    }
}

const DEFAULT_R_MAX: f64 = 2.0;
const DEFAULT_R_MIN: f64 = -0.15;
const DEFAULT_R_FREQ: usize = 1;
const DEFAULT_R_PHASE: f64 = 0.0;
const DEFAULT_M_MAG: f64 = 0.5;
const DEFAULT_M_FREQ: usize = 1;
const DEFUALT_M_PHASE: f64 = 0.0;

#[derive(Clone, Debug)]
pub struct StochasticBehavior {
    pub r_max: f64,
    pub r_min: f64,
    pub r_freq: usize,
    pub r_phase: f64,
    pub m_mag: f64,
    pub m_freq: usize,
    pub m_phase: f64,
}

impl StochasticBehavior {
    pub fn new() -> Self {
        Self {
            r_max: DEFAULT_R_MAX,
            r_min: DEFAULT_R_MIN,
            r_freq: DEFAULT_R_FREQ,
            r_phase: DEFAULT_R_PHASE,
            m_mag: DEFAULT_M_MAG,
            m_freq: DEFAULT_M_FREQ,
            m_phase: DEFUALT_M_PHASE,
        }
    }

    pub fn random_vector_behavior(
        &mut self,
        max_mag: f64,
        min_mag: f64,
        frequency: usize,
        phase: f64,
    ) {
        assert!(
            min_mag < max_mag,
            "Minimum magnitude must be smaller than maximum magnitude!"
        );
        assert_eq!(frequency % 2, 1, "Frequency must be an odd number!");
        assert!(phase > 0.0, "Phase must be a posative number!");

        self.r_max = max_mag;
        self.r_min = min_mag;
        self.r_freq = frequency;
        self.r_phase = phase;
    }

    pub fn momentum_behavior(&mut self, magnitude: f64, frequency: usize, phase: f64) {
        assert_eq!(frequency % 2, 1, "Frequency must be an odd number!");
        assert!(phase > 0.0, "Phase must be a posative number!");

        self.m_mag = magnitude;
        self.m_freq = frequency;
        self.m_phase = phase;
    }
}

impl fmt::Display for StochasticBehavior {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(
            f,
            "Random Vector Properties \n \t max: {:.5}, min:{:.5}, phase: {:.5}, frequency: {} \n",
            self.r_min, self.r_max, self.r_phase, self.r_freq
        )?;
        write!(
            f,
            "Momentum Properties \n \t magnitude: {:.5}, phase: {:.5}, frequency: {}",
            self.m_mag, self.m_phase, self.m_freq
        )
    }
}
